An implementation of David King and John Launchbury's 
SCC algorithm from "Structuring depth-first algorithms in Haskell"

> module SCC (depordering) where

We pass off the imperative-like bit (see section 5.2 of the paper)
to a separate module to allow for differing implementations. 
The base module contains datatypes needed by those implementations.

> import SCC_Base
> import SCC_nonST_Prune

> import Array
> import List
> import Maybe
> import FiniteMaps

The concept of the  interface following is to some extent lifted/adapted 
from the GHC source code:

> type DataNode k i = ([k],[k],i)

> depordering :: Ord k => [DataNode k i] -> [[i]]

> depordering nodes = 
>     let (graph, decoder) = encodeGraph $ collect $ sortBy sourcecmp nodes
>     in map (concat . map decoder . postorder) $ scc' graph
>          where sourcecmp (sks1,_,_) (sks2,_,_) = compare sks1 sks2
>                collect [] = []
>                collect ((sks1,tks1,info1):nodes) 
>                  = case collect nodes of
>                     [] -> [(sks1,tks1,[info1])]
>                     (sks2,tks2,infos2):nodes' 
>                       -> if sks1==sks2 
>                          then (sks1,tks1++tks2,info1:infos2):nodes'
>                          else (sks1,tks1,[info1]):(sks2,tks2,infos2):nodes'

> type GK = Int

> encodeGraph :: Ord k => [DataNode k i] -> (Graph GK, GK -> i)

> encodeGraph nodes = (keysgraph, \key -> nodesmap!key)
>     where bounds = (0,length nodes-1) :: (GK,GK)
>           nodesmap = listArray bounds $ map (\(_,_,info) -> info) nodes
>           numberednodes = zip [0..] nodes
>           sourcekeys = map (\(n,(keys,_,_)) -> (keys,n)) numberednodes
>           keysmap = listToFM $ concat $ map expand sourcekeys
>           expand (keys,n) = map (\key -> (key,n)) keys
>           makeedge (n,(sourcekeys,targetkeys,_))
>              = (n,catMaybes $ map (lookupFM keysmap) targetkeys)
>           keysgraph = array bounds $ map makeedge numberednodes

The depth-first search/SCC code itself

> type Table v a = Array v a
> type Graph v = Table v [v]

> vertices :: Ix v => Graph v -> [v]
> vertices = indices

> type Edge v = (v,v)

> edges :: Ix v => Graph v -> [Edge v]
> edges g = [ (v,w) | v <- vertices g, w <- g!v]

> mapT :: Ix v => (v -> a -> b) -> Table v a -> Table v b
> mapT f t = array (bounds t) [(v, f v (t!v)) | v <- indices t]


> buildG :: Ix v => Bounds v -> [Edge v] -> Graph v
> buildG bnds es = accumArray (flip (:)) [] bnds es

> transposeG :: Ix v => Graph v -> Graph v
> transposeG g = buildG (bounds g) (reverseE g)

> reverseE :: Ix v => Graph v -> [Edge v]
> reverseE g = [ (w,v) | (v,w) <- edges g]

> generate :: Ix v => Graph v -> v -> Tree v
> generate g v = Node v (map (generate g) (g!v))

> dfs :: Ix v => Graph v -> [v] -> Forest v
> dfs g vs = prune (bounds g) (map (generate g) vs)

> dff :: Ix v => Graph v -> Forest v
> dff g = dfs g (vertices g)

> postorder :: Tree a -> [a]
> postorder (Node a ts) = postorderF ts ++ [a]

> postorderF :: Forest a -> [a]
> postorderF ts = concat (map postorder ts)

> postOrd :: Ix v => Graph v -> [v]
> postOrd g = postorderF (dff g)

> scc' :: Ix v => Graph v -> Forest v
> scc' g = dfs g (reverse (postOrd (transposeG g)))